"""Just pass the OCR process."""

from typing import Tuple


def ocr(path, language=["auto"], options: dict = None) -> Tuple[str, bool]:
    return "", True


def config(language: str = None) -> dict:
    return {}


def get() -> dict:
    return None



import re
from typing import Tuple, Callable
import httpx
import os
from ..Doc2X.Exception import nomal_retry
import concurrent.futures


def get_imgcdnlink_list(text: str) -> Tuple[list, list]:
    """
    Extract the image links from the text.
    Return the image links list and the image path list.
    """
    patterns = [
        (r'<img\s+src="([^"]+)"\s+alt="([^"]*)">', lambda m: (m.group(0), m.group(1))),
        (
            r'<img\s+style="[^"]*"\s+src="([^"]+)"\s*/>',
            lambda m: (m.group(0), m.group(1)),
        ),
        (r'<img\s+src="([^"]+)"\s*/>', lambda m: (m.group(0), m.group(1))),
        (r"!\[[^\]]*\]\(<([^>]+)>\)", lambda m: (m.group(0), m.group(1))),
        (r"!\[[^\]]*\]\(([^)]+)\)", lambda m: (m.group(0), m.group(1))),
    ]

    origin_text_list = []
    imgpath_list = []

    for pattern, extract in patterns:
        for match in re.finditer(pattern, text):
            origin_text, src = extract(match)
            if origin_text not in origin_text_list:  # 检查是否已存在于列表中
                origin_text_list.append(origin_text)
                imgpath_list.append(src)

    return origin_text_list, imgpath_list


@nomal_retry()
def download_img_from_url(url: str, savepath: str) -> None:
    """
    Download the image from the url to the savepath, changing the extension based on the content type.
    """
    with httpx.stream("GET", url) as response:
        content_type = response.headers.get("Content-Type")
        if content_type:
            extension = content_type.split("/")[-1]
            savepath = f"{savepath}.{extension}"
        with open(savepath, "wb") as file:
            for chunk in response.iter_bytes():
                file.write(chunk)
        return extension


def md_replace_imgs(
    mdfile: str,
    replace,
    skip: str = None,
    outputpath: str = "",
    relative: bool = False,
    threads: int = 5,
) -> bool:
    """Replace the image links in the markdown file (cdn links -> local file).

    Args:
        mdfile (str): The markdown file path.
        replace: Str or function to replace the image links. For str only "local" accepted. Defaults to "local".
        skip (str, optional): The URL start with this string will be skipped. Defaults to None. For example, "https://menghuan1918.github.io/pdfdeal-docs".
        outputpath (str, optional): The output path to save the images, if not set, will create a folder named as same as the markdown file name and add `_img`. **⚠️Only works when `replace` is "local".**
        relative (bool, optional): The output path to save the images with relative path. Defaults to False. **⚠️Only works when `replace` is "local".**
        threads (int, optional): The number of threads to download the images. Defaults to 5.

    Returns:
        bool: If all images are downloaded successfully, return True, else return False.
    """
    if isinstance(replace, str) and replace == "local":
        pass
    elif isinstance(replace, Callable):
        pass
    else:
        raise ValueError("The replace must be 'local' or a function.")

    with open(mdfile, "r", encoding="utf-8") as file:
        content = file.read()

    imglist, imgpath = get_imgcdnlink_list(content)
    if len(imglist) == 0:
        print("No image links found in the markdown file.")
        return True

    no_outputppath_flag = False
    if outputpath == "":
        no_outputppath_flag = True
        outputpath = os.path.splitext(mdfile)[0] + "_img"
    os.makedirs(outputpath, exist_ok=True)

    print(f"Start to download images from file {mdfile}")

    def download_image(i, imgurl, outputpath, relative, mdfile):
        if not imgurl.startswith("http"):
            print(f"\nNot a valid url: {imgurl}, Skip it.")
            return None
        elif skip and imgurl.startswith(skip):
            print(f"\nSkip the image: {imgurl}, because it starts with {skip}.")
            return None
        try:
            savepath = f"{outputpath}/img{i}"
            extension = download_img_from_url(imgurl, savepath)
            savepath = f"{savepath}.{extension}"
            if relative:
                savepath = os.path.relpath(savepath, os.path.dirname(mdfile))
                return (imglist[i], f"![{imgurl}](<{savepath}>)\n")
            else:
                savepath = os.path.abspath(savepath)
                return (imglist[i], f"![{imgurl}](<{savepath}>)\n")
        except Exception as e:
            print(
                f"Error to download the image: {imgurl}, continue to download the next image:\n {e}"
            )
            return None

    replacements = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        futures = [
            executor.submit(
                download_image,
                i,
                imgurl,
                outputpath,
                relative,
                mdfile,
            )
            for i, imgurl in enumerate(imgpath)
        ]
        for future in concurrent.futures.as_completed(futures):
            result = future.result()
            if result:
                replacements.append(result)

    flag = True
    for old, new in replacements:
        content = content.replace(old, new)

    if len(replacements) < len(imglist):
        print("Some images may not be downloaded successfully. Please check the log.")
        flag = False

    if isinstance(replace, Callable):
        imglist, imgpath = get_imgcdnlink_list(content)

        @nomal_retry()
        def upload_task(i, img_path, replace):
            if img_path.startswith("http://") or img_path.startswith("https://"):
                print(f"Skip the image: {img_path}, because it is a url.")
                return None, None, None
            try:
                remote_file_name = f"{os.path.splitext(os.path.basename(mdfile))[0]}_{os.path.basename(img_path)}"
                new_url, flag = replace(img_path, remote_file_name)
                if flag:
                    img_url = f"![{os.path.splitext(os.path.basename(mdfile))[0]}](<{new_url}>)\n"
                    return img_url, True, i
                else:
                    print(f"=====\nError to upload the image: {img_path}, {new_url}")
                    print("Continue to upload the next image.")
                    return new_url, False, i
            except Exception as e:
                print(f"=====\nError to upload the image: {img_path}, {e}")
                print("Continue to upload the next image.")
                return new_url, False, i

        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
            futures = [
                executor.submit(upload_task, i, img_path, replace)
                for i, img_path in enumerate(imgpath)
            ]
            for future in concurrent.futures.as_completed(futures):
                new_url, flag, i = future.result()
                if flag:
                    content = content.replace(imglist[i], new_url)
                elif flag is None:
                    pass
                else:
                    print(f"=====\nError to upload the image: {imgpath[i]}, {new_url}")
                    print("Continue to upload the next image.")
                    flag = False

        for img in imgpath:
            os.remove(img)

        if no_outputppath_flag:
            try:
                os.rmdir(outputpath)
            except Exception as e:
                print(f"\nError to remove the folder: {outputpath}, {e}")

    with open(mdfile, "w", encoding="utf-8") as file:
        file.write(content)

    print(f"Finish to process images in file {mdfile}.")
    return flag


def mds_replace_imgs(
    path: str,
    replace,
    outputpath: str = "",
    relative: bool = False,
    skip: str = None,
    threads: int = 2,
    down_load_threads: int = 3,
) -> Tuple[list, list, bool]:
    """Replace the image links in the markdown file (cdn links -> local file).

    Args:
        path (str): The markdown file path.
        replace: Str or function to replace the image links. For str only "local" accepted. Defaults to "local".
        outputpath (str, optional): The output path to save the images, if not set, will create a folder named as same as the markdown file name and add `_img`.  **⚠️Only works when `replace` is "local".**
        relative (bool, optional): Whether to save the images with relative path. Defaults to False, **⚠️Only works when `replace` is "local".**
        skip (str, optional): The URL start with this string will be skipped. Defaults to None. For example, "https://menghuan1918.github.io/pdfdeal-docs".
        threads (int, optional): The number of threads to download the images. Defaults to 2.
        down_load_threads (int, optional): The number of threads to download the images in one md file. Defaults to 3.

    Returns:
        Tuple[list, list, bool]:
                `list1`: list of successfilly processed md file path, if the file failed, its path will be empty string
                `list2`: list of failed files's error message and its original file path, if some files are successful, its error message will be empty string
                `bool`: If all files are processed successfully, return True, else return False.
    """
    if isinstance(replace, str) and replace == "local":
        pass
    elif isinstance(replace, Callable):
        pass
    else:
        raise ValueError("The replace must be 'local' or a function.")

    from pdfdeal import gen_folder_list

    mdfiles = gen_folder_list(path=path, mode="md", recursive=True)
    if len(mdfiles) == 0:
        print("No markdown file found in the path.")
        return [], [], True

    import concurrent.futures

    def process_mdfile(mdfile, replace, outputpath, relative):
        try:
            md_replace_imgs(
                mdfile=mdfile,
                replace=replace,
                outputpath=outputpath,
                relative=relative,
                skip=skip,
                threads=down_load_threads,
            )
            return mdfile, None
        except Exception as e:
            return mdfile, e

    success_files = []
    fail_files = []
    Fail_flag = True

    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        futures = [
            executor.submit(process_mdfile, mdfile, replace, outputpath, relative)
            for mdfile in mdfiles
        ]
        for future in concurrent.futures.as_completed(futures):
            mdfile, error = future.result()
            if error:
                Fail_flag = False
                fail_files.append({"error": str(error), "path": mdfile})
                print(f"Error to process the markdown file: {mdfile}, {error}")
                print("Continue to process the next markdown file.")
            else:
                success_files.append(mdfile)

    print(
        f"\n[MARKDOWN REPLACE] Successfully processed {len(success_files)}/{len(mdfiles)} markdown files."
    )

    if Fail_flag is False:
        print("Some markdown files process failed.")
        return success_files, fail_files, True

    return success_files, fail_files, False



from pdfdeal import Doc2X
from pdfdeal.file_tools import get_files
import os


def test_single_pic2file():
    client = Doc2X()
    filepath, _, _ = client.pic2file(
        image_file="tests/image/sample.png",
        output_path="./Output/test/single/pic2file",
        output_names=["pic_sample1.docx"],
        output_format="docx",
    )
    if filepath[0] != "":
        assert os.path.exists(filepath[0])
        assert os.path.isfile(filepath[0])
        assert filepath[0].endswith(".docx")
        assert os.path.basename(filepath[0]) == "pic_sample1.docx"


def test_multiple_pic2file():
    client = Doc2X()
    file_list, rename = get_files("tests/image", "img", "docx")
    success, failed, flag = client.pic2file(
        image_file=file_list,
        output_path="./Output/test/multiple/pic2file",
        output_names=rename,
        output_format="docx",
    )
    assert flag
    assert len(success) == len(failed) == 3
    for s in success:
        if s != "":
            assert s.endswith("sample1.docx") or s.endswith("sample.docx")


# def test_multiple_high_rpm():
#     client = Doc2X()
#     file_list = ["tests/image/sample.png" for _ in range(20)]
#     success, failed, flag = client.pic2file(
#         image_file=file_list,
#         output_path="./Output/test",
#     )
#     assert len(success) == len(failed) == 20
#     i = 0
#     for s in success:
#         if s != "":
#             assert s.endswith(".zip")
#         else:
#             i += 1
#     print(f"===Failed {i} times===")



import io
from PIL import Image
from pypdf import PdfReader
import re
import emoji
import unicodedata
import os
import zipfile
import shutil
from typing import Tuple
from ..Doc2X.Types import Support_File_Type, OutputFormat
from .dealmd import split_of_md
import traceback


def clean_text(text):
    """
    Clean the text
    """
    # remove extra whitespaces
    text = re.sub(r"\n\s*\n", "\n\n", text)

    # remove special characters
    text = re.sub(r"-\d+-", "", text)

    def is_valid_unicode(char):
        # Check if a character is a valid unicode character
        try:
            unicodedata.name(char)
            return True
        except ValueError:
            return False

    cleaned_text = []
    for char in text:
        if (
            char.isalnum()
            or char.isspace()
            or emoji.is_emoji(char)
            or is_valid_unicode(char)
        ):
            cleaned_text.append(char)
    text = "".join(cleaned_text)

    return text


def clear_cache():
    """
    Clear the cache folder
    """
    temp_image_folder = os.path.join(
        os.path.expanduser("~"), ".cache", "pdfdeal", "pictures"
    )
    for file in os.listdir(temp_image_folder):
        file_path = os.path.join(temp_image_folder, file)
        if os.path.isfile(file_path):
            os.remove(file_path)


def extract_text_and_images(pdf_path, ocr, language=["ch_sim", "en"], GPU=False):
    """
    Extract text and images from a PDF file
    """
    Text = []

    # Open the PDF file
    with open(pdf_path, "rb") as file:
        reader = PdfReader(file)

        for page in reader.pages:
            # Get the text content of the page
            text = page.extract_text()
            temp_image_folder = os.path.join(
                os.path.expanduser("~"), ".cache", "pdfdeal", "pictures"
            )
            os.makedirs(temp_image_folder, exist_ok=True)
            clear_cache()
            # Get the images on the page
            images = page.images
            for id, image in enumerate(images):
                image_data = image.data
                image_stream = io.BytesIO(image_data)
                pil_image = Image.open(image_stream)
                # Save to HOME/.cache/pdfdeal/pictures, if the directory does not exist, create it
                temp_image_path = os.path.join(
                    os.path.expanduser("~"),
                    ".cache",
                    "pdfdeal",
                    "pictures",
                    f"{id}.png",
                )
                pil_image.save(temp_image_path)
            option = {"GPU": GPU}
            # Use ocr to extract text from images
            ocr_text, All_Done = ocr(temp_image_folder, language, option)
            text += f"\n{ocr_text}"
            Text.append(clean_text(text))
        clear_cache()
    return Text, All_Done


def gen_folder_list(path: str, mode: str, recursive: bool = False) -> list:
    """Generate a list of all files in the folder

    Args:
        path (str): The path of the folder to be processed
        mode (str): The type of file to find, 'pdf', 'img' or 'md'
        recursive (bool): Whether to search subdirectories recursively

    Raises:
        ValueError: If the mode is not 'pdf', 'img' or 'md'

    Returns:
        list: The list of full paths of the files
    """
    if os.path.isfile(path):
        raise ValueError("The input should be a folder.")

    def _find_files(path, mode):
        if mode == "pdf":
            return [
                os.path.join(path, f) for f in os.listdir(path) if f.endswith(".pdf")
            ]
        elif mode == "img":
            return [
                os.path.join(path, f)
                for f in os.listdir(path)
                if f.endswith(".png") or f.endswith(".jpg") or f.endswith(".jpeg")
            ]
        elif mode == "md":
            return [
                os.path.join(path, f) for f in os.listdir(path) if f.endswith(".md")
            ]
        else:
            raise ValueError("Mode should be 'pdf','img' or 'md'")

    if not recursive:
        return _find_files(path, mode)
    else:
        all_files = []
        for root, _, files in os.walk(path):
            all_files.extend(_find_files(root, mode))
        return all_files


def get_files(path: str, mode: str, out: str) -> Tuple[list, list]:
    """Generate a list of files in the folder, keeps the structure of the file the same before and after processing

    Args:
        path (str): The path of the folder to be processed
        mode (str): Which type of file to process, 'pdf' or 'img'
        out (str): Which type of file want to output, `md`, `md_dollar`, `latex` or `docx`, or `pdf` if you are using for RAG

    Returns:
        Tuple[list, list]: The list of full paths and relative paths, use in (like)`input` and `output_format`
    """
    # check if input is a file or a folder
    if os.path.isfile(path):
        raise ValueError("The input should be a folder.")
    mode = Support_File_Type(mode)
    if isinstance(mode, Support_File_Type):
        mode = mode.value
    if out != "pdf":
        out = OutputFormat(out)
        if isinstance(out, OutputFormat):
            out = out.value
        out = out if out == "docx" else "zip"
    full_paths = []
    relative_paths = []

    if mode == "pdf":
        extensions = [".pdf"]
    elif mode == "img":
        extensions = [".jpg", ".jpeg", ".png"]

    if os.path.isfile(path):
        if any(path.lower().endswith(ext) for ext in extensions):
            full_paths.append(path)
            rel_path = os.path.relpath(path, os.path.dirname(path))
            rel_path_out = os.path.splitext(rel_path)[0] + "." + out
            relative_paths.append(rel_path_out)
            return full_paths, relative_paths

    for root, dirs, files in os.walk(path):
        for file in files:
            if any(file.lower().endswith(ext) for ext in extensions):
                full_path = os.path.join(root, file)
                full_paths.append(full_path)

                rel_path = os.path.relpath(full_path, path)
                rel_path_out = os.path.splitext(rel_path)[0] + "." + out
                relative_paths.append(rel_path_out)

    return full_paths, relative_paths


def unzip(zip_path: str, rename: bool = True) -> str:
    """Unzip the zip file and return the path of the extracted folder

    Args:
        zip_path (str): The path to the zip file
        rename (bool, optional): If rename the .md or .tex file with the unziped folder name. Defaults to True.

    Raises:
        Exception: If the zip file is not valid

    Returns:
        str: The path of the extracted folder
    """
    folder_name = os.path.splitext(os.path.basename(zip_path))[0]
    extract_path = os.path.join(os.path.dirname(zip_path), folder_name)
    os.makedirs(extract_path, exist_ok=True)
    try:
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(extract_path)
        os.remove(zip_path)
        if rename:
            # Only find the first .md or .tex file and rename it
            for root, _, files in os.walk(extract_path):
                for file in files:
                    if file.endswith(".md") or file.endswith(".tex"):
                        old_file_path = os.path.join(root, file)
                        new_file_path = os.path.join(
                            root, folder_name + os.path.splitext(file)[1]
                        )
                        os.rename(old_file_path, new_file_path)
                        return extract_path
        return extract_path
    except Exception as e:
        raise Exception(f"Unzip file error! {e}")


def texts_to_file(texts, filepath, output_format="txt"):
    """
    Write texts to a file.
    `texts`: a list of strings, each string is a paragraph.
    `filepath`: the folder path of the file to write.
    `output_format`: the format of the output file, default is "txt",acceptable values are "txt", "md".

    Return the path of the output file.
    """
    import time

    if output_format not in ["txt", "md"]:
        raise ValueError("The output format is not supported.")
    if not os.path.exists(filepath):
        os.makedirs(filepath)
    try:
        temp_file = os.path.join(filepath, f"{time.time()}.{output_format}")
    except Exception as e:
        raise RuntimeError(f"Error: {e}")
    with open(temp_file, "w") as file:
        for text in texts:
            file.write(text)
            file.write("\n")
    return temp_file


def list_rename(files: list, new_name: list) -> list:
    """
    Rename files according to the new name list

    Args:
        `files`: list of files
        `new_name`: list of new names

    Return:
        `list`: list of new files
    """
    if len(files) != len(new_name):
        raise ValueError("The length of files and new_name should be the same.")
    new_files = []
    for file, name in zip(files, new_name):
        if file == "":
            new_files.append("")
            continue
        new_file = os.path.join(os.path.dirname(file), name)
        new_file_folder = os.path.dirname(new_file)
        os.makedirs(new_file_folder, exist_ok=True)
        if os.path.dirname(file) != new_file_folder:
            shutil.move(file, new_file_folder)
        os.rename(os.path.join(new_file_folder, os.path.basename(file)), new_file)
        new_files.append(new_file)
    return new_files


def auto_split_md(
    mdfile: str,
    mode: str = "title",
    out_type: str = "single",
    split_str: str = "=+=+=+=+=+=+=+=+=",
    output_path: str = "./Output",
) -> Tuple[str, bool]:
    """Split the md file

    Args:
        mdfile (str): The path to md file
        mode (str, optional): The way to split. **Only support `title`(split by every title) now.** Defaults to "title".
        out_type (str, optional): The way to output the splited file. Only support `single`(one file) and `replace`(replace the original file) now. Defaults to "single".
        split_str (str, optional): The string to split the md file. Defaults to `=+=+=+=+=+=+=+=+=`.
        output_path (str, optional): The path to output the splited file. Defaults to "./Output". Not work when `out_type` is `replace`.

    Returns:
        Tuple[str,bool] : The path to the output file and whether the file is splited.
    """
    if not os.path.exists(mdfile):
        raise FileNotFoundError(f"The file {mdfile} does not exist.")
    elif os.path.isdir(mdfile):
        raise IsADirectoryError(f"The path {mdfile} is a directory.")

    #! In the future, will support more modes.
    try:
        new_content = split_of_md(mdfile=mdfile, mode="title")
    except Exception as e:
        print(traceback.format_exc())
        print(f"=====\nError deal with {mdfile} : {e}")
        return f"Error deal with {mdfile} : {e}", False

    write_contene = ""
    for content in new_content:
        write_contene += split_str + "\n" + content + "\n"

    if out_type == "replace":
        with open(mdfile, "w", encoding="utf-8") as file:
            file.writelines(write_contene)
        return mdfile, True

    elif out_type == "single":
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        new_file = os.path.join(output_path, os.path.basename(mdfile))
        with open(new_file, "w", encoding="utf-8") as file:
            file.writelines(write_contene)
        return new_file, True


def auto_split_mds(
    mdpath: str,
    mode: str = "title",
    out_type: str = "single",
    split_str: str = "=+=+=+=+=+=+=+=+=",
    output_path: str = "./Output",
    recursive: bool = True,
) -> Tuple[list, list, bool]:
    """Split the md files in the folder

    Args:
        mdpath (str): The path to the folder containing md files
        mode (str, optional): The way to split. **Only support `title`(split by every title) now.** Defaults to "title".
        out_type (str, optional): The way to output the splited file. Only support `single`(one file) and `replace`(replace the original file) now. Defaults to "single".
        split_str (str, optional): The string to split the md file. Defaults to `=+=+=+=+=+=+=+=+=`.
        output_path (str, optional): The path to output the splited file. Defaults to "./Output". Not work when `out_type` is `replace`.
        recursive (bool, optional): Whether to search subdirectories recursively. Defaults to True.

    Returns:
        Tuple[list,list,str]:
        will return `list1`,`list2`,`bool`
        `list1`: is the list of the output files, if some files are not splited, the element will be `""`
        `list2`: is the list of the error message and its original file path, if some files are successfully splited, the element will be `""`
        `bool`: True means that at least one file process failed
    """
    if not os.path.exists(mdpath):
        raise FileNotFoundError(f"The path {mdpath} does not exist.")
    elif os.path.isfile(mdpath):
        raise IsADirectoryError(f"The path {mdpath} is a file.")

    md_files = gen_folder_list(mdpath, mode="md", recursive=recursive)
    if len(md_files) == 0:
        return [], False
    success = []
    failed = []
    flag = False
    for mdfile in md_files:
        try:
            temp, is_splited = auto_split_md(
                mdfile=mdfile,
                mode=mode,
                out_type=out_type,
                split_str=split_str,
                output_path=output_path,
            )
            if is_splited:
                success.append(temp)
                failed.append({"error": "", "file": ""})
            else:
                success.append("")
                failed.append({"error": temp, "file": mdfile})
                flag = True
        except Exception as e:
            success.append("")
            failed.append({"error": e, "file": mdfile})
            flag = True
    print(
        f"MD SPLIT: {sum([1 for i in success if i != ''])}/{len(success)} files are successfully splited."
    )
    print(f"Note the split string is :\n{split_str}")
    if flag:
        for failed_file in failed:
            if failed_file["error"] != "":
                print(
                    f"=====\nError deal with {failed_file['file']} : {failed_file['error']}"
                )
    return success, failed, flag



"""Connect to Doc2X OCR API"""

from PIL import Image
import os
from typing import Tuple

LANGUAGES = ["简体中文", "Enlish"]
WORDS_CN = [
    "📇 请输入 Doc2X 的身份令牌，个人用户请访问 https://doc2x.noedgeai.com/ 获取：",
    "⚠️ 验证 Doc2X 的身份令牌失败，请检查网络连接或者身份令牌是否正确",
    "📌 请选择 Doc2X 的速率限制，含意为同时请求数量，建议输入 A 以自动选择速率限制：",
]
WORDS_EN = [
    "📇 Please enter the API key of the Doc2X, for personal use, visit https://doc2x.com/ to get the key:",
    "⚠️ Failed to verify the API key of Doc2X, please check the network connection or the API key",
    "📌 Please select the rate limit of Doc2X, means number of simultaneous requests, it is recommended to enter A to automatically select the rate limit:",
]
WORDS = [WORDS_CN, WORDS_EN]


def doc2x_judgements(image_file):
    """
    Whether the image is samll enough to be enable purely formulaic model
    """
    with Image.open(image_file) as img:
        size = img.size
    if size[0] < 50 and size[1] < 50:
        return 1
    else:
        return 0


def ocr(path, language=["auto"], options: dict = None) -> Tuple[str, bool]:
    """Do OCR with Doc2X

    Args:
        path (str): The path to pictures folder or a picture
        language (list, optional): No need for Doc2X
        options (dict, must): Need `{"Doc2X_Key": key, "Doc2X_RPM": RPM}`

    Raises:
        Exception: If the Doc2X limit is 0
        Exception: If the number of files in the folder exceeds the limit
        Exception: If the input is invalid

    Returns:
        Tuple[str, bool]: The OCR text and if all the files are done
    """
    from pdfdeal import Doc2X

    api_key = options["api_key"]
    rpm = options.get("rpm", None)
    if rpm is None:
        Client = Doc2X(apikey=api_key)
    else:
        Client = Doc2X(apikey=api_key, thread=rpm)

    try:
        limit = Client.get_limit()
    except Exception as e:
        raise Exception(f"Get error! {e}")
    if limit == 0:
        raise Exception("The Doc2X limit is 0, please check your account.")

    text = ""
    All_Done = True
    if os.path.isfile(path) and path.endswith((".jpg", ".png", ".jpeg")):
        try:
            equation = doc2x_judgements(image_file=path)
            texts, Failed, Fail_flag = Client.pic2file(
                image_file=path,
                output_format="txts",
                equation=equation,
            )
            for t in texts:
                text += t + "\n"
            if Fail_flag:
                for fail in Failed:
                    if fail["error"] != "":
                        print(f"Get error when using Doc2X to do ocr. {fail['error']}")
                All_Done = False
        except Exception as e:
            print(f"Get error when using Doc2X to do ocr and pass to next file. {e}")
            All_Done = False
            pass
    elif os.path.isdir(path):
        file_count = 0
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.endswith((".jpg", ".png", ".jpeg")):
                    file_count += 1
        if file_count > limit:
            raise Exception(
                f"The number of files in the folder exceeds the limit, please check the folder: {path}"
            )

        for root, dirs, files in os.walk(path):
            for file in files:
                file_path = os.path.join(root, file)
                if file_path.endswith((".jpg", ".png", ".jpeg")):
                    try:
                        # * Since the size and dimensions of each image may be different, batch processing mode is not used
                        equation = doc2x_judgements(image_file=file_path)
                        texts, Failed, Fail_flag = Client.pic2file(
                            image_file=path,
                            output_format="txts",
                            equation=equation,
                        )
                        for t in texts:
                            text += t + "\n"
                        if Fail_flag:
                            for fail in Failed:
                                if fail["error"] != "":
                                    print(
                                        f"Get error when using Doc2X to do ocr. {fail['error']}"
                                    )
                            All_Done = False
                    except Exception as e:
                        print(
                            f"Get error when using Doc2X to do ocr and pass to next file. {e}"
                        )
                        All_Done = False
                        pass
    return text, All_Done


def config(language: str = None) -> dict:
    """Set the configuration of the local directory"""
    from ...Watch.config import curses_select

    if language is None:
        language = curses_select(LANGUAGES, "Please select the language:")
    words = WORDS[language]
    key = input(words[0])
    from pdfdeal import Doc2X

    try:
        Doc2X(apikey=key)
    except Exception as e:
        raise Exception(f"{words[1]}:\n {e}")
    RPM = input(words[2])
    assert RPM.isdigit() or RPM == "A" or RPM == "a", "The input is invalid."
    if RPM == "A" or RPM == "a":
        if key.startswith("sk-"):
            RPM = 10
        else:
            RPM = 1
    return {"Doc2X_Key": key, "Doc2X_RPM": int(RPM)}


def get(config: dict) -> dict:
    """Get the option setting from the configuration"""
    try:
        return {"api_key": config["Doc2X_Key"], "rpm": config["Doc2X_RPM"]}
    except KeyError:
        raise KeyError("The configuration is invalid, please check the configuration")



import asyncio
from functools import wraps
import time
import traceback
import sys
from concurrent.futures import ThreadPoolExecutor


class RateLimit(Exception):
    """
    Error when rate limit is reached.
    """

    pass


class RequestError(Exception):
    """
    Error when request is not successful, usually because of the file broken.
    """

    pass


class FileError(Exception):
    """
    Error when file is not found or cannot be opened or othes.
    """

    pass


def async_retry(max_retries=2, backoff_factor=2):
    """
    Decorator to retry an async function when an exception is raised.
    `max_retries`: Maximum number of retries.
    `backoff_factor`: Factor to increase the wait time between retries.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            retries = 0
            while retries < max_retries:
                try:
                    return await func(*args, **kwargs)
                except RateLimit:
                    raise RateLimit
                except FileError as e:
                    print("Error details:\n")
                    print(traceback.format_exc())
                    print("\n===================\n")
                    raise FileError(e)
                except RequestError as e:
                    print("Error details:\n")
                    print(traceback.format_exc())
                    print("\n===================\n")
                    raise RequestError(f"{e} \nThis usually means the file is broken.")
                except Exception as e:
                    last_exception = e
                    wait_time = backoff_factor**retries
                    print("\n===================\n")
                    print(f"⚠️Get exception {e}. \n♻️Retrying in {wait_time} seconds.")
                    await asyncio.sleep(wait_time)
                    retries += 1
            print("\n===================\n")
            print("Error details:\n")
            print(traceback.format_exc())
            print("\n===================\n")
            raise last_exception

        return wrapper

    return decorator


def nomal_retry(max_retries=3, backoff_factor=2):
    """
    Decorator to retry an async function when an exception is raised.
    `max_retries`: Maximum number of retries.
    `backoff_factor`: Factor to increase the wait time between retries.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            retries = 0
            while retries < max_retries:
                try:
                    return func(*args, **kwargs)
                except RateLimit:
                    raise RateLimit
                except FileError as e:
                    print("\n===================\n")
                    print("Error details:\n")
                    print(traceback.format_exc())
                    print("\n===================\n")
                    raise e
                except Exception as e:
                    last_exception = e
                    wait_time = backoff_factor**retries
                    print("\n===================\n")
                    print(f"Get exception {e}. \n Retrying in {wait_time} seconds.")
                    time.sleep(wait_time)
                    retries += 1
            print("\n===================\n")
            print("Error details:\n")
            print(traceback.format_exc())
            print("\n===================\n")
            raise last_exception

        return wrapper

    return decorator


def run_async(coro):
    """This function is used to run async function in sync way. As `asyncio.run` not work in jupyter notebook.

    Args:
        coro (_type_): The function to run.

    Returns:
        _type_: The result of the function.
    """
    if "IPython" in sys.modules:
        # Jupyter Notebook
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            with ThreadPoolExecutor() as executor:
                future = executor.submit(asyncio.run, coro)
                return future.result()
    # Python
    return asyncio.run(coro)



import httpx
import json
import os
import re
from typing import Tuple, Literal
from .Exception import RateLimit, FileError, RequestError, async_retry

Base_URL = "https://api.doc2x.noedgeai.com/api"


@async_retry()
async def refresh_key(key: str) -> str:
    """Get the real key by the personal key

    Args:
        key (str): The personal key

    Raises:
        Exception: Failed to verify key

    Returns:
        str: The real key
    """
    url = f"{Base_URL}/token/refresh"
    timeout = httpx.Timeout(30)
    async with httpx.AsyncClient(timeout=timeout) as client:
        get_res = await client.post(url, headers={"Authorization": "Bearer " + key})
    if get_res.status_code == 200:
        try:
            return json.loads(get_res.content.decode("utf-8"))["data"]["token"]
        except Exception as e:
            raise Exception(f"Failed to verify key: {e}")
    else:
        raise Exception(f"Failed to verify key: {get_res.status_code}:{get_res.text}")


async def check_folder(path: str) -> bool:
    """Make sure the path is a folder

    Args:
        path (str): The path to check

    Raises:
        Exception: Input path already exists as a file
        Exception: Input path is not a directory
        Exception: Create folder error

    Returns:
        bool: Whether the path is a folder
    """
    try:
        os.makedirs(path, exist_ok=True)
        assert os.path.isdir(path)
    except FileExistsError:
        raise Exception("Input path already exists as a file!")
    except NotADirectoryError:
        raise Exception("Input path is not a directory!")
    except Exception as e:
        raise Exception(f"Create folder error! {e}")
    return True


@async_retry()
async def uuid2file(
    apikey: str,
    uuid: str,
    output_format: Literal["md", "md_dollar", "latex", "docx"],
    output_path: str = "./Output",
) -> str:
    """Get the file by the uuid

    Args:
        apikey (str): The key
        uuid (str): The uuid of the file
        output_format (Literal[&quot;md&quot;, &quot;md_dollar&quot;, &quot;latex&quot;, &quot;docx&quot;]): The output format
        output_path (str, optional): The output path. Defaults to "./Output".

    Raises:
        Exception: Input path is not a directory
        RateLimit: Rate limit exceeded
        Exception: Download file error

    Returns:
        str: The path of the file
    """
    await check_folder(output_path)
    url = f"{Base_URL}/export?request_id={uuid}&to={output_format}"
    download_format = output_format if output_format == "docx" else "zip"
    timeout = httpx.Timeout(120)
    async with httpx.AsyncClient(timeout=timeout) as client:
        get_res = await client.get(url, headers={"Authorization": "Bearer " + apikey})
    if get_res.status_code == 200:
        output_path = os.path.join(output_path, uuid + "." + download_format)
        try:
            with open(output_path, "wb") as f:
                f.write(get_res.content)
            return output_path
        except Exception as e:
            raise Exception(f"Dowload file error! {e}")
    else:
        if get_res.status_code == 429:
            raise RateLimit()
        elif get_res.status_code == 400:
            raise RequestError(get_res.text)
        else:
            raise Exception(
                f"Download file error! {get_res.status_code}:{get_res.text}"
            )


@async_retry()
async def get_limit(apikey: str) -> int:
    """Get the limit of the key

    Args:
        apikey (str): The key

    Raises:
        RuntimeError: The key is invalid

    Returns:
        int: The limit of the key
    """
    if apikey.startswith("sk-"):
        url = f"{Base_URL}/v1/limit"
    else:
        url = f"{Base_URL}/platform/limit"
    async with httpx.AsyncClient() as client:
        get_res = await client.get(url, headers={"Authorization": "Bearer " + apikey})
    if get_res.status_code == 200:
        try:
            return int(get_res.json()["data"]["remain"])
        except Exception as e:
            raise RuntimeError(
                f"Get limit error with {e}! {get_res.status_code}:{get_res.text}"
            )
    else:
        raise RuntimeError(f"Get limit error! {get_res.status_code}:{get_res.text}")


@async_retry()
async def upload_pdf(
    apikey: str,
    pdffile: str,
    ocr: bool = True,
    translate: bool = False,
    language: str = "zh",
    model: str = "deepseek",
) -> str:
    """Upload pdf file to server and return the uuid of the file

    Args:
        apikey (str): The key
        pdffile (str): The pdf file path
        ocr (bool, optional): Do OCR or not. Defaults to True.
        translate (bool, optional): Do translate or not. Defaults to False.
        language (str, optional): The language of the file. Defaults to "zh", only valid when translate is True.
        model (str, optional): The model of the file. Defaults to "deepseek", only valid when translate is True.

    Raises:
        FileError: Input file size is too large
        FileError: Open file error
        RateLimit: Rate limit exceeded
        Exception: Upload file error

    Returns:
        str: The uuid of the file
    """
    if apikey.startswith("sk-"):
        url = f"{Base_URL}/v1/async/pdf"
    else:
        url = f"{Base_URL}/platform/async/pdf"
    try:
        file = {"file": open(pdffile, "rb")}
        if os.path.getsize(pdffile) > 100 * 1024 * 1024:
            raise FileError("PDF dile size should be less than 100MB!")
    except Exception as e:
        raise FileError(f"Open file error! {e}")
    ocr = 1 if ocr else 0
    translate = 2 if translate else 1
    timeout = httpx.Timeout(120)
    if translate == 1:
        async with httpx.AsyncClient(timeout=timeout) as client:
            post_res = await client.post(
                url,
                headers={"Authorization": "Bearer " + apikey},
                files=file,
                data={"ocr": ocr, "parse_to": translate},
            )
    else:
        async with httpx.AsyncClient(timeout=timeout) as client:
            post_res = await client.post(
                url,
                headers={"Authorization": "Bearer " + apikey},
                files=file,
                data={
                    "ocr": ocr,
                    "parse_to": translate,
                    "lang": language,
                    "model": model,
                },
            )
    if post_res.status_code == 200:
        try:
            if (
                "parse_task_limit_exceeded"
                == json.loads(post_res.content.decode("utf-8"))["code"]
            ):
                raise RateLimit()
            else:
                return json.loads(post_res.content.decode("utf-8"))["data"]["uuid"]
        except RateLimit:
            raise RateLimit()
        except Exception as e:
            raise Exception(
                f"Upload file error with {e} ! {post_res.status_code}:{post_res.text}"
            )
    elif post_res.status_code == 429:
        raise RateLimit()
    elif post_res.status_code == 400:
        raise RequestError(post_res.text)
    else:
        raise Exception(f"Upload file error! {post_res.status_code}:{post_res.text}")


@async_retry()
async def upload_img(
    apikey: str,
    imgfile: str,
    formula: bool = False,
    img_correction: bool = False,
) -> str:
    """Upload image file to server and return the uuid of the file

    Args:
        apikey (str): The key
        imgfile (str): The image file path
        formula (bool, optional): Only formula or not. Defaults to False.
        img_correction (bool, optional): Do img correction or not. Defaults to False.

    Raises:
        FileError: Image file size is too large
        FileError: Open file error
        RateLimit: Rate limit exceeded
        Exception: Upload file error

    Returns:
        str: The uuid of the file
    """
    if apikey.startswith("sk-"):
        url = f"{Base_URL}/v1/async/img"
    else:
        url = f"{Base_URL}/platform/async/img"
    formula = 1 if formula else 0
    img_correction = 1 if img_correction else 0
    try:
        file = {"file": open(imgfile, "rb")}
        if os.path.getsize(imgfile) > 10 * 1024 * 1024:
            raise FileError("Image file size should be less than 10MB!")
    except Exception as e:
        raise FileError(f"Open file error! {e}")
    timeout = httpx.Timeout(120)
    async with httpx.AsyncClient(timeout=timeout) as client:
        post_res = await client.post(
            url,
            headers={"Authorization": "Bearer " + apikey},
            files=file,
            data={"equation": formula, "img_correction": img_correction},
        )
    if post_res.status_code == 200:
        try:
            if (
                "parse_task_limit_exceeded"
                == json.loads(post_res.content.decode("utf-8"))["code"]
            ):
                raise RateLimit()
            else:
                return json.loads(post_res.content.decode("utf-8"))["data"]["uuid"]
        except RateLimit:
            raise RateLimit()
        except Exception as e:
            raise Exception(
                f"Upload file error with {e}! {post_res.status_code}:{post_res.text}"
            )
    elif post_res.status_code == 429:
        raise RateLimit()
    elif post_res.status_code == 400:
        raise RequestError(post_res.text)
    else:
        raise Exception(f"Upload file error! {post_res.status_code}:{post_res.text}")


async def decode_data(datas: json, convert: bool) -> Tuple[list, list]:
    """Decode the data

    Args:
        datas (json): The data
        convert (bool): Convert "[" and "[[" to "$" and "$$"

    Returns:
        Tuple[list, list]: The texts and locations
    """
    texts = []
    locations = []
    for data in datas["result"]["pages"]:
        try:
            text = data["md"]
        except KeyError:
            text = ""
        if convert:
            text = re.sub(r"\\[()]", "$", text)
            text = re.sub(r"\\[\[\]]", "$$", text)
        try:
            url = data["url"]
        except KeyError:
            url = ""
        # When processing image, there is no page_idx
        try:
            page_id = data["page_idx"]
        except KeyError:
            page_id = 0
        location = {
            "url": url,
            "page_idx": page_id,
            "page_width": data["page_width"],
            "page_height": data["page_height"],
        }
        texts.append(text)
        locations.append(location)
    return texts, locations


async def decode_translate(datas: json, convert: bool) -> Tuple[list, list]:
    """Decode the translate data

    Args:
        datas (json): The data
        convert (bool): Convert "[" and "[[" to "$" and "$$"

    Returns:
        Tuple[list, list]: The texts and locations
    """
    texts = []
    locations = []
    Rawinput = json.loads(datas["result"])
    for data in Rawinput:
        try:
            text = data["raw"]
            translate_text = data["translated"]
            if convert:
                text = re.sub(r"\\[()]", "$", text)
                text = re.sub(r"\\[\[\]]", "$$", text)
                translate_text = re.sub(r"\\[()]", "$", translate_text)
                translate_text = re.sub(r"\\[\[\]]", "$$", translate_text)
            location = {
                "raw_text": text,
                "page_idx": data["page_idx"],
                "page_width": data["page_width"],
                "page_height": data["page_height"],
                "x": data["x"],
                "y": data["y"],
            }
            texts.append(translate_text)
            locations.append(location)
        except KeyError:
            continue
    return texts, locations


@async_retry()
async def uuid_status(
    apikey: str,
    uuid: str,
    convert: bool = False,
    translate: bool = False,
) -> Tuple[int, str, list]:
    """Get the status of the file

    Args:
        apikey (str): The key
        uuid (str): The uuid of the file
        convert (bool, optional): Convert or not. Defaults to False.
        translate (bool, optional): Translate or not. Defaults to False.

    Raises:
        RuntimeError: Pages limit exceeded
        RuntimeError: Unknown status
        Exception: Get status error

    Returns:
        Tuple[int, str, list]: The progress, status and texts
    """
    if apikey.startswith("sk-"):
        url = f"{Base_URL}/v1/async/status?uuid={uuid}"
    else:
        url = f"{Base_URL}/platform/async/status?uuid={uuid}"
    if translate:
        url += "&parse_to=2"
    timeout = httpx.Timeout(30)
    async with httpx.AsyncClient(timeout=timeout) as client:
        get_res = await client.get(url, headers={"Authorization": "Bearer " + apikey})
    if get_res.status_code == 200:
        datas = None
        try:
            datas = json.loads(get_res.content.decode("utf-8"))["data"]
        except Exception as e:
            raise Exception(
                f"Get status error with {e}! {get_res.status_code}:{get_res.text}"
            )
        if datas["status"] == "ready":
            return 0, "Waiting for processing", [], []

        elif datas["status"] == "processing":
            return int(datas["progress"]), "Processing file", [], []

        elif datas["status"] == "translate_processing":
            return int(datas["progress"]), "Translating file", [], []

        elif datas["status"] == "success":
            texts, locations = await decode_data(datas, convert)
            return 100, "Success", texts, locations

        elif datas["status"] == "translate_success":
            texts, locations = await decode_translate(datas, convert)
            return 100, "Translate success", texts, locations

        elif datas["status"] == "pages limit exceeded":
            raise RuntimeError("Pages limit exceeded!")

        else:
            raise RuntimeError(f"Unknown status! {get_res.text}")

    raise Exception(f"Get status error! {get_res.status_code}:{get_res.text}")


async def process_status(original_file: list, output_file: list):
    """Check the status of the files, error or success

    Args:
        original_file (list): The original file list
        output_file (list): The output file list

    Returns:
        _type_: The success file list, error file list and has error flag
    """
    success_file = []
    error_file = []

    for orig, out in zip(original_file, output_file):
        # if the output type is texts or in translate mode, the output is a list
        if isinstance(out, list):
            success_file.append(out)
            error_file.append({"error": "", "path": ""})
        elif isinstance(out, dict):
            success_file.append(out)
            error_file.append({"error": "", "path": ""})
        elif out.startswith("Error"):
            success_file.append("")
            error_file.append({"error": out, "path": orig})
        else:
            success_file.append(out)
            error_file.append({"error": "", "path": ""})
    try:
        has_error_flag = any(file.startswith("Error") for file in output_file)
    except AttributeError:
        has_error_flag = False

    return success_file, error_file, has_error_flag



"""Connect to local tesseract OCR"""

import os
from typing import Tuple


def ocr(path, language=["eng"], options: dict = None) -> Tuple[str, bool]:
    """Do OCR with tesseract

    Args:
        path (str): The path to pictures folder or a picture
        language (list, optional): The language use for tesseract, only first one will use. Defaults to ["eng"].
        options (dict, optional): No need for tesseract. Defaults to None.

    Raises:
        ImportError: If not install `pytesseract`

    Returns:
        Tuple[str, bool]:  The OCR text and if all the files are done
    """
    try:
        import pytesseract
    except ImportError:
        raise ImportError(
            "Please install pytesseract first, use 'pip install pytesseract'"
        )
    text = ""
    All_Done = True
    if os.path.isfile(path) and path.endswith((".jpg", ".png", ".jpeg")):
        try:
            text += pytesseract.image_to_string(path, lang=language[0])
            text += "\n"
        except Exception as e:
            print(
                f"Get error when using pytesseract to do ocr and pass to next file. {e}"
            )
            All_Done = False
            pass
    elif os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            for file in files:
                file_path = os.path.join(root, file)
                if file_path.endswith((".jpg", ".png", ".jpeg")):
                    try:
                        text += pytesseract.image_to_string(file_path, lang=language[0])
                        text += "\n"
                    except Exception as e:
                        print(
                            f"Get error when using pytesseract to do ocr and pass to next file. {e}"
                        )
                        All_Done = False
                        pass
    return text, All_Done


def config(language: str = None) -> dict:
    return {}


def get() -> dict:
    return None



"""Connect to a local directory"""

import logging
from typing import Tuple
import shutil
import os

LANGUAGES = ["简体中文", "Enlish"]
WORDS_CN = [
    "📂 请输入目输出的目标文件路径：",
    "📦 目录不存在，是否创建？[Y/n]：",
    "⚠️ 目标必须是一个文件夹",
    "⚠️ 没有目标文件夹的写入权限，请确保您有写入权限",
    "⚠️ 目录不为空，请输入一个空目录",
]
WORDS_EN = [
    "📂 Please enter the target file path:",
    "📦 The directory does not exist, do you want to create it? [Y/n]:",
    "⚠️ The target must be a directory",
    "⚠️ No write permission for the target directory, please make sure you have write permission",
    "⚠️ The directory is not empty, please enter an empty directory",
]
WORDS = [WORDS_CN, WORDS_EN]


def connect(file_list: list, base_path: str, options: dict) -> Tuple[list, list, bool]:
    """Connect to a local directory"""
    target_path = options["target_path"]
    success = []
    faied = []
    flag = False
    for file in file_list:
        try:
            shutil.move(os.path.join(base_path, file), os.path.join(target_path, file))
            success.append(file)
            faied.append({"error": "", "path": ""})
            logging.info(f"Succeed to move {file} to {target_path}")
        except Exception as e:
            success.append("")
            faied.append({"error": str(e), "path": file})
            flag = True
            logging.error(f"Failed to move {file} to {target_path}, error: {str(e)}")
    return success, faied, flag


def config(language: str = None) -> dict:
    """Set the configuration of the local directory"""
    from ..config import curses_select

    if language is None:
        language = curses_select(LANGUAGES, "Please select the language:")
    words = WORDS[language]
    while True:
        target_path = input(words[0])
        if not os.path.exists(target_path):
            if input(words[1]).lower() != "n":
                os.makedirs(target_path)
                break
        else:
            # Check the target path
            if not os.path.isdir(target_path):
                raise NotADirectoryError(words[2])
            # Check the permission of the target path
            try:
                with open(os.path.join(target_path, "test"), "w"):
                    pass
                os.remove(os.path.join(target_path, "test"))
            except PermissionError:
                raise PermissionError(words[3])
            # Check the content of the target path
            if os.listdir(target_path):
                raise FileExistsError(words[4])
    return {"target_path": os.path.abspath(target_path)}


def get(config: dict) -> str:
    """Get the option setting from the configuration"""
    try:
        return {"target_path": config["target_path"]}
    except KeyError:
        raise KeyError("The configuration is invalid, please check the configuration")


from pdfdeal import Doc2X
from pdfdeal.file_tools import get_files
import os


def test_pdfdeal():
    client = Doc2X()
    filepath, _, _ = client.pdfdeal(
        pdf_file="tests/pdf/sample.pdf",
        output_path="./Output/test/pdfdeal",
    )
    if filepath[0] != "":
        assert os.path.exists(filepath[0])
        assert os.path.isfile(filepath[0])
        assert filepath[0].endswith(".pdf")


def test_multiple_pdfdeal():
    client = Doc2X()
    file_list, rename = get_files("tests/pdf", "pdf", "pdf")
    success, failed, flag = client.pdfdeal(
        pdf_file=file_list,
        output_path="./Output/test/multiple/pdfdeal",
        output_names=rename,
    )
    assert flag
    assert len(success) == len(failed) == 3
    for s in success:
        if s != "":
            assert s.endswith(".pdf")
    for f in failed:
        if f["path"] != "":
            assert f["path"].endswith("sample_bad.pdf")


# def test_multiple_high_rpm_v2():
#     client = Doc2X()
#     file_list = ["tests/pdf/sample.pdf" for _ in range(20)]
#     success, failed, flag = client.pdfdeal(
#         input=file_list,
#         path="./Output/test/high_rpm/pdfdeal",
#     )
#     assert len(success) == len(failed) == 20
#     i = 0
#     for s in success:
#         if s != "":
#             assert os.path.exists(s)
#             assert s.endswith(".pdf")
#         else:
#             i += 1
#     print(f"===Failed {i} times===")



from pdfdeal import Doc2X
import pytest


def test_client():
    client = Doc2X()
    assert client is not None

def test_client_with_invalid_key():
    with pytest.raises(Exception):
        Doc2X(apikey="invalid_key")


def test_client_thread():
    client = Doc2X(thread=50)
    assert client is not None


def test_client_unsupported_args():
    with pytest.raises(ValueError):
        Doc2X(rpm=1,thread=2)


"""Modular pdf file recognition engine"""

from typing import Tuple

BUILD_IN_TOOL = ["doc2x_pdf"]

def load_build_in_tool(tool: str):
    """
    Load the build-in tool engine

    Args:
        tool: The name of the tool engine, support `doc2x_pdf`

    Returns:
        tool, config, get
    """
    tool_mapping = {
        "doc2x_pdf": Tool_doc2x
    }
    tool_init =  tool_mapping.get(tool)
    return tool_init()


def Tool_doc2x() -> Tuple[list, list, bool]:
    """
    deal pdf file with Doc2X

    Returns:
        tool, config, get
    """
    from .Tool.doc2x import tool, config, get

    return tool, config, get



"""Modular ocr engines"""

from typing import Tuple

BUILD_IN_OCR = ["doc2x_ocr", "easyocr", "pytesseract"]


def load_build_in_ocr(ocr: str):
    """
    Load the build-in OCR engine

    Args:
        ocr: The name of the OCR engine, support `doc2x_ocr`, `easyocr`, `pytesseract`, `pass`

    Returns:
        ocr, config, get
    """
    ocr_mapping = {
        "doc2x_ocr": OCR_pass,
        "easyocr": OCR_easyocr,
        "pytesseract": OCR_pytesseract,
    }
    ocr_init = ocr_mapping.get(ocr, OCR_pass)
    return ocr_init()


def OCR_easyocr() -> Tuple[str, bool]:
    """
    OCR with easyocr

    Returns:
        ocr, config, get
    """
    from .OCR.easyocr import ocr, config, get

    return ocr, config, get


def OCR_pytesseract():
    """
    OCR with pytesseract

    Returns:
        ocr, config, get
    """
    from .OCR.pytesseract import ocr, config, get

    return ocr, config, get


def OCR_pass():
    """
    Pass the OCR process

    Returns:
        ocr, config, get
    """
    from .OCR.passocr import ocr, config, get

    return ocr, config, get


def Doc2X_OCR():
    """
    OCR with Doc2X, `ocr` function's `options` args need a dict `{"Doc2X_Key": key, "Doc2X_RPM": int(RPM)}`

    Returns:
        ocr, config, get
    """
    from .OCR.doc2x import ocr, config, get

    return ocr, config, get



import argparse
from pdfdeal.Watch.store import (
    get_global_setting,
    change_one_global_setting,
    delete_one_global_setting,
)
import os
from pdfdeal import Doc2X
from pdfdeal.file_tools import get_files
from pdfdeal.Watch.config import curses_select

LANGUAGES = ["简体中文", "Enlish"]
WORDS_CN = [
    "📇 请输入 Doc2X 的身份令牌，个人用户请访问 https://doc2x.noedgeai.com/ 获取, 将会自动保存至本地：",
    "⚠️ 验证 Doc2X 的身份令牌失败，请检查网络连接或者身份令牌是否正确",
    "📌 请选择 Doc2X 的速率限制，含意为同时请求数量，强烈建议输入 A 以自动选择速率限制：",
    "🚧 请选择要处理的文件类型:",
    "📂 请输入要处理的文件或文件夹：",
    "⚠️ 未找到所在文件或文件夹",
]
WORDS_EN = [
    "📇 Please enter the API key of the Doc2X, for personal use, visit https://doc2x.com/ to get the key, will auto save to local:",
    "⚠️ Failed to verify the API key of Doc2X, please check the network connection or the API key",
    "📌 Please select the rate limit of Doc2X, means number of simultaneous requests, it is recommended to enter A to automatically select the rate limit:",
    "🚧 Please select the type of file to process:",
    "📂 Please enter the file or folder to deal with:",
    "⚠️ The file or folder does not exist",
]
WORDS = [WORDS_CN, WORDS_EN]


def i18n(language):
    """Maybe the lazy i18n solution, but it works."""
    if language is None:
        language = curses_select(LANGUAGES, "Please select the language:")
    return WORDS[language], language


def set_doc2x_key(language):
    """Set the Doc2X key and rate limit."""
    words, language = i18n(language)
    key = input(words[0])
    try:
        Doc2X(apikey=key)
    except Exception as e:
        raise Exception(f"{words[1]}:\n {e}")
    RPM = input(words[2])
    assert RPM.isdigit() or RPM == "A" or RPM == "a", "The input is invalid."
    if RPM == "A" or RPM == "a":
        if key.startswith("sk-"):
            RPM = 10
        else:
            RPM = 1
    return {"Doc2X_Key": key, "Doc2X_RPM": int(RPM)}, language


def get_file_folder(language):
    words, language = i18n(language)
    while True:
        file_folder = input(words[4])
        if os.path.exists(file_folder):
            break
        print(words[5])
    return file_folder, language


def file_type(language):
    words, language = i18n(language)
    file_type = curses_select(selects=["PDF", "Picture"], show=words[3])
    if file_type == 0:
        return False, True, language
    else:
        return True, False, language


def main():
    parser = argparse.ArgumentParser(
        description="Using doc2x to deal with pictures or pdfs"
    )
    parser.add_argument("filename", help="PDF or picture file/folder", nargs="?")
    parser.add_argument(
        "-y",
        help="Will skip any scenarios that require a second input from the user.",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "-k",
        "--api_key",
        help="The API key of Doc2X, if not set, will use the global setting",
        required=False,
    )
    parser.add_argument(
        "-r",
        "--rpm",
        help="The rate limit of Doc2X, DO NOT set if you don't know",
        required=False,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="The output folder of the result, if not set, will set to './Output'",
        required=False,
    )
    parser.add_argument(
        "-f",
        "--format",
        help="The output format of the result, accept md、md_dollar、latex、docx, default is md_dollar",
        required=False,
        choices=["md", "md_dollar", "latex", "docx"],
    )
    parser.add_argument(
        "-i",
        "--image",
        help="If the input is a picture, set this flag to True, or will ask you",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "-p",
        "--pdf",
        help="If the input is a pdf, set this flag to True, or will ask you",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "--equation",
        help="Whether to use the equation model, only works for pictures, default is False",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--clear",
        help="Clear all the global setting about Doc2X",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "--graphrag",
        help="Change md document to txt form, used for output is converted to the txt form accepted by graphRAG. The output format needs to be md or md_dollar at this time",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "--unzip",
        help="Unzip the output file, only works for the output is zip file",
        required=False,
        action="store_true",
    )
    # Only if need user input, will ask language
    language = None
    args = parser.parse_args()
    rpm = None

    if args.clear:
        delete_one_global_setting("Doc2X_Key")
        delete_one_global_setting("Doc2X_RPM")
        print("Clear all the global setting.")

    if args.api_key is None:
        try:
            api_key = str(get_global_setting()["Doc2X_Key"])
            rpm = int(get_global_setting()["Doc2X_RPM"])
            print("Find API: ", api_key[:5] + "*" * (len(api_key) - 10) + api_key[-5:])
        except Exception:
            if args.y:
                pass
            else:
                api_key = None
                print(
                    "The global setting does not exist, please set the global setting first."
                )
                doc2x_setting, language = set_doc2x_key(language)
                for key, value in doc2x_setting.items():
                    change_one_global_setting(key, value)
                api_key = str(doc2x_setting["Doc2X_Key"])
                rpm = int(doc2x_setting["Doc2X_RPM"])
    else:
        api_key = str(args.api_key)

    if rpm is None:
        rpm = int(args.rpm) if args.rpm else 10 if api_key.startswith("sk-") else 1

    image = args.image
    pdf = args.pdf
    if not image and not pdf:
        if args.y:
            pdf = True
        else:
            image, pdf, language = file_type(language)
    if image and pdf:
        raise ValueError("You can only choose one type of file to process.")

    filename = args.filename

    if filename is None:
        if args.y:
            filename = "./"
        else:
            filename, language = get_file_folder(language)

    output = args.output if args.output else "./Output"

    format = args.format if args.format else "md_dollar"

    equation = args.equation

    if api_key is None or api_key == "":
        Client = Doc2X()
    else:
        Client = Doc2X(apikey=api_key, thread=rpm)

    if args.graphrag:
        assert format in [
            "md",
            "md_dollar",
        ], "The output format needs to be md or md_dollar at this time"

    if image:
        files, rename = get_files(filename, "img", format)
        success, fail, flag = Client.pic2file(
            image_file=files,
            output_path=output,
            output_names=rename,
            output_format=format,
            equation=equation,
        )

    if pdf:
        files, rename = get_files(filename, "pdf", format)
        success, fail, flag = Client.pdf2file(
            pdf_file=files,
            output_path=output,
            output_names=rename,
            output_format=format,
        )

    for file in success:
        if file != "":
            file = os.path.abspath(file)
            print(f"Save to: {file}")
    if flag:
        print("Some files failed to process, please check the error message.")
        print("Try to save the failed file path and reasons to fail.txt")
        try:
            with open("fail.txt", "w") as f:
                for file in fail:
                    if file["path"] != "":
                        f.write(str(file))
        except Exception as e:
            print(f"Failed to save the failed files to fail.txt: {e}")
            print("The failed files are:")
            for file in fail:
                print(file)

    if args.graphrag or args.unzip:
        from pdfdeal.FileTools.file_tools import unzip

        for file in success:
            if file != "" and file.endswith(".zip"):
                file = os.path.abspath(file)
                try:
                    unzip(zip_path=file, rename=True)
                except Exception as e:
                    print(f"Failed to unzip the file: {file}, error: {e}")
        output_folder = os.path.abspath(output)
        if args.graphrag:
            for root, dirs, files in os.walk(output_folder):
                for file in files:
                    if file.endswith(".md"):
                        file_path = os.path.join(root, file)
                        new_file_path = file_path[:-3] + ".txt"
                        os.rename(file_path, new_file_path)
        print(f"Unzip and rename the files in {output_folder} successfully.")

    print(f"===\nLeft Doc2X pages: {Client.get_limit()}\n Have a nice day!")


if __name__ == "__main__":
    main()



import os
from typing import Tuple

LANGUAGES = ["简体中文", "Enlish"]
WORDS_CN = ["🔌 是否使用GPU加速？注意easyocr仅支持CUDA加速："]
WORDS_EN = [
    "🔌 Do you want to use GPU acceleration? Note that easyocr only supports CUDA acceleration:"
]
WORDS = [WORDS_CN, WORDS_EN]


def ocr(path, language=["auto"], options: dict = None) -> Tuple[str, bool]:
    """Do OCR with easyocr

    Args:
        path (str): The path to pictures folder or a picture
        language (list, optional): Language use. Defaults to ["auto"].
        options (dict, optional): Need `{"GPU": bool}`. Defaults is `{"GPU": False}`

    Raises:
        ImportError: If not install `easyocr`

    Returns:
        Tuple[str, bool]: The OCR text and if all the files are done
    """
    try:
        import easyocr
    except ImportError:
        raise ImportError("Please install easyocr first, use 'pip install easyocr'")
    GPU = options.get("gpu", False)
    reader = easyocr.Reader(language, gpu=GPU)
    All_Done = True
    texts = ""
    if os.path.isfile(path) and path.endswith((".jpg", ".png", ".jpeg")):
        try:
            result = reader.readtext(path, detail=0, paragraph=True)
        except Exception as e:
            result = [""]
            All_Done = False
            print(f"Get error when using easyocr to do ocr and pass to next file. {e}")
        texts += "\n".join(result)
    elif os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            for file in files:
                file_path = os.path.join(root, file)
                if file_path.endswith((".jpg", ".png", ".jpeg")):
                    try:
                        result = reader.readtext(file_path, detail=0, paragraph=True)
                    except Exception as e:
                        result = [""]
                        All_Done = False
                        print(
                            f"Get error when using easyocr to do ocr and pass to next file. {e}"
                        )
                    texts += "\n".join(result)
    return texts, All_Done


def config(language: str = None) -> dict:
    """Get the config of easyocr"""
    from ...Watch.config import curses_select

    if language is None:
        language = curses_select(LANGUAGES, "Please select the language:")
    words = WORDS[language]
    gpu = curses_select(["❎ No", "✅ Yes"], words[0])
    if gpu == "✅ Yes":
        return {"GPU": True}
    return {"GPU": False}


def get(config: dict) -> dict:
    """Get the config of easyocr"""
    try:
        return {"GPU": config["GPU"]}
    except KeyError:
        raise KeyError("The configuration is invalid, please check the configuration")



import os
from .file_tools import extract_text_and_images
from .ocr import load_build_in_ocr, BUILD_IN_OCR
from reportlab.lib.pagesizes import letter
from reportlab.pdfgen import canvas
from reportlab.pdfbase.cidfonts import UnicodeCIDFont
from reportlab.pdfbase import pdfmetrics
from ..Doc2X.Types import RAG_OutputType
import uuid
from typing import Tuple,Callable
from .file_tools import list_rename


def strore_pdf(pdf_path, Text):
    c = canvas.Canvas(pdf_path, pagesize=letter)

    pdfmetrics.registerFont(UnicodeCIDFont("STSong-Light"))
    c.setFont("STSong-Light", 12)

    for text in Text:
        lines = text.split("\n")
        for i, line in enumerate(lines):
            c.setFont("STSong-Light", 12)
            c.drawString(100, 750 - i * 13, line)
        c.showPage()
    c.save()


def deal_pdf_back(
    input,
    output: RAG_OutputType = RAG_OutputType.PDF,
    ocr=None,
    language: list = ["ch_sim", "en"],
    GPU: bool = False,
    path: str = "./Output",
):
    pdf_path = input

    Text, All_Done = extract_text_and_images(
        pdf_path=pdf_path, ocr=ocr, language=language, GPU=GPU
    )

    if output == "texts":
        return Text, All_Done

    else:
        filename = pdf_path.split("/")[-1].replace(".pdf", f"_{uuid.uuid4()}.{output}")
        os.makedirs(path, exist_ok=True)
        if output == "md":
            try:
                output_md_path = os.path.join(path, filename)
                with open(output_md_path, "w") as file:
                    for text in Text:
                        file.write(text + "\n")
                return output_md_path, All_Done
            except Exception as e:
                RuntimeError(f"Failed to save the markdown file: {e}")
        elif output == "pdf":
            output_pdf_path = os.path.join(path, filename)
            try:
                strore_pdf(output_pdf_path, Text)
                return output_pdf_path, All_Done
            except Exception as e:
                RuntimeError(f"Failed to save the PDF file: {e}")


def deal_pdf(
    pdf_file,
    output_format: str = "pdf",
    output_names: list = None,
    ocr=None,
    language: list = ["ch_sim", "en"],
    GPU: bool = False,
    output_path: str = "./Output",
    option: dict = {},
) -> Tuple[list, list, bool]:
    """Deal with PDF files with OCR, make PDF more readable for RAG

    Args:
        pdf_file (str or list): input file path, str or list
        output_format (str, optional): the type of output, "texts" "md" or "pdf", default is "pdf". Defaults to "pdf".
        output_names (list, optional): Custom Output File Names, must be the same length as `image_file`. Defaults to None.
        ocr (function, optional): custom ocr/tool function, not define will use easyocr. Or use `string`: `pytesseract` to use pytesseract, string `pass` to skip OCR. Defaults to None.
        language (list, optional): the language used in OCR, default is ["ch_sim", "en"] for easyocr, ["eng"] for pytesseract. Defaults to ["ch_sim", "en"].
        GPU (bool, optional): whether to use GPU in OCR, default is False, not working for pytesseract. Defaults to False.
        output_path (str, optional): the path of folder to save the output, default is "./Output", only used when output is "md" or "pdf". Defaults to "./Output".
        option (dict, optional): the option of the OCR/tool. Defaults to `{}`.

    Returns:
        tuple[list,list,str]:
        will return `list1`,`list2`,`bool`
            `list1`: list of successful files path, if some files are failed, its path will be empty string
            `list2`: list of failed files's error message and its original file path, id some files are successful, its error message will be empty string
            `bool`: True means that at least one file process failed
    """
    output_format = RAG_OutputType(output_format)
    if isinstance(output_format, RAG_OutputType):
        output_format = output_format.value

    if isinstance(pdf_file, str):
        if not pdf_file.endswith(".pdf"):
            RuntimeError("The input must be path to a PDF file")
        pdf_file = [pdf_file]

    ocr = ocr or "easyocr"

    if isinstance(ocr, Callable):
        if len(ocr.__code__.co_varnames) == 2:
            # if using tool function
            option["output"] = output_path
            return ocr(pdf_file, option)

    elif isinstance(ocr, str):
        if ocr in BUILD_IN_OCR:
            ocr, _, _ = load_build_in_ocr(ocr)
        else:
            RuntimeError(f"OCR {ocr} is not supported.")
    else:
        raise RuntimeError("The ocr must be a function or a string")

    success_file = []
    failed_file = []
    error_flag = False
    for pdf_path in pdf_file:
        try:
            output, All_Done = deal_pdf_back(
                pdf_path, output_format, ocr, language, GPU, output_path
            )
            success_file.append(output)
            if not All_Done:
                failed_file.append(
                    {
                        "error": "Some pictures are failed to OCR, but the text and reset pictures is extracted",
                        "file": pdf_path,
                    }
                )
            else:
                failed_file.append({"error": "", "file": ""})
        except Exception as e:
            All_Done = True  # * As this flag is use to determine if the OCR is done, we set it to True here
            success_file.append("")
            failed_file.append({"error": str(e), "file": pdf_path})
            error_flag = True
    print(
        f"PDFDEAL Progress: {sum(1 for s in success_file if s != '')}/{len(pdf_file)} files successfully processed."
    )
    if All_Done is False:
        print(
            "Some pictures are failed to OCR, but the text and reset pictures is extracted"
        )
    if error_flag:
        for f in failed_file:
            if f["error"] != "":
                print(
                    f"-----\nFailed to process file: {f['file']} with error: {f['error']}\n-----"
                )

    if output_names is not None:
        success_file = list_rename(success_file, output_names)

    return success_file, failed_file, error_flag



from .FileTools.dealpdfs import deal_pdf
from .FileTools.extract_img import md_replace_imgs
from .doc2x import Doc2X
from .FileTools.file_tools import gen_folder_list, get_files

__all__ = ['deal_pdf', 'md_replace_imgs', 'Doc2X', 'gen_folder_list', 'get_files']


"""Connect to Doc2X API to deal with pdf file."""

from typing import Tuple

LANGUAGES = ["简体中文", "Enlish"]
WORDS_CN = [
    "📇 请输入 Doc2X 的身份令牌，个人用户请访问 https://doc2x.noedgeai.com/ 获取：",
    "⚠️ 验证 Doc2X 的身份令牌失败，请检查网络连接或者身份令牌是否正确",
    "📌 请选择 Doc2X 的速率限制，含意为同时请求数量，建议输入 A 以自动选择速率限制：",
]
WORDS_EN = [
    "📇 Please enter the API key of the Doc2X, for personal use, visit https://doc2x.com/ to get the key:",
    "⚠️ Failed to verify the API key of Doc2X, please check the network connection or the API key",
    "📌 Please select the rate limit of Doc2X, means number of simultaneous requests, it is recommended to enter A to automatically select the rate limit:",
]
WORDS = [WORDS_CN, WORDS_EN]


def tool(path: str, options: dict) -> Tuple[list, list, bool]:
    """
    deal pdf file with Doc2X
    """
    from pdfdeal import Doc2X

    api_key = options["api_key"]
    rpm = options.get("rpm", None)
    if rpm is None:
        Client = Doc2X(apikey=api_key)
    else:
        Client = Doc2X(apikey=api_key, thread=rpm)

    try:
        limit = Client.get_limit()
    except Exception as e:
        raise Exception(f"Get error! {e}")
    if limit == 0:
        raise Exception("The Doc2X limit is 0, please check your account.")

    return Client.pdfdeal(pdf_file=path, output_path=options["output"], convert=False)


def config(language: str = None) -> dict:
    """Set the configuration of the local directory"""
    from ...Watch.config import curses_select

    if language is None:
        language = curses_select(LANGUAGES, "Please select the language:")
    words = WORDS[language]
    key = input(words[0])
    from pdfdeal import Doc2X

    try:
        Doc2X(apikey=key)
    except Exception as e:
        raise Exception(f"{words[1]}:\n {e}")
    RPM = input(words[2])
    assert RPM.isdigit() or RPM == "A" or RPM == "a", "The input is invalid."
    if RPM == "A" or RPM == "a":
        if key.startswith("sk-"):
            RPM = 10
        else:
            RPM = 1
    return {"Doc2X_Key": key, "Doc2X_RPM": int(RPM)}


def get(config: dict) -> dict:
    """Get the option setting from the configuration"""
    try:
        return {"api_key": config["Doc2X_Key"], "rpm": config["Doc2X_RPM"]}
    except KeyError:
        raise KeyError("The configuration is invalid, please check the configuration")



"""The main part of deal with the pdf files"""

import logging
import os

from ..FileTools.ocr import BUILD_IN_OCR, load_build_in_ocr
from ..FileTools.tool import BUILD_IN_TOOL, load_build_in_tool
from .connect import BUILD_IN_CONNECT


def tools(tool: str, file_list: list, config: dict):
    #! Need to rewrite the dealpdf function
    temp_path = os.path.join(os.path.expanduser("~"), ".cache", "pdfdeal", "dealed")
    #! Not support the custom options now
    if tool not in BUILD_IN_TOOL:
        raise Exception(f"Tool {tool} is not supported.")
    tool_init_func = load_build_in_tool(tool)
    if tool == "doc2x_pdf":
        from pdfdeal import Doc2X

        Client = Doc2X(apikey=config["Doc2X_Key"], thread=config["Doc2X_RPM"])
        tool_func = tool_init_func(Client)
    options = {"output": temp_path}
    return tool_func(file_list, options)


def ocrs(ocr: str, config: dict):
    if ocr not in BUILD_IN_OCR:
        raise Exception(f"OCR {ocr} is not supported.")
    ocr_init_func = load_build_in_ocr(ocr)
    if ocr == "doc2x_ocr":
        from pdfdeal import Doc2X

        Client = Doc2X(apikey=config["Doc2X_Key"], thread=config["Doc2X_RPM"])
        return ocr_init_func(Client)
    return ocr_init_func


def option_part(tool_or_ocr: str, file_list: list, config: dict):
    """Do the main process of the program, do ocr or file tools.

    Args:
        tool_or_ocr (str): The name of the tool or ocr
        file_list (list): The list of the file path to deal with
        config (dict): The configuration of the program

    Returns:
        _type_: _description_
    """
    if tool_or_ocr in BUILD_IN_OCR:
        #! Need to rewrite the dealpdf function
        from pdfdeal import deal_pdf

        temp_path = os.path.join(os.path.expanduser("~"), ".cache", "pdfdeal", "dealed")
        ocr = ocrs(tool_or_ocr, config)
        success, fail, flag = deal_pdf(file_list, ocr=ocr, output_path=temp_path)
    else:
        success, fail, flag = tools(tool_or_ocr, file_list, config)
    return success, fail, flag


def main_process(file_list: list, config: dict):
    done_files = []
    success, fail, flag = option_part(config["option1"], file_list, config)
    if flag:
        logging.error("The first option failed, fallback to the second option.")
        done_files.extend([file for file in success if file != ""])
        failed_files = [file['path'] for file in fail]
        success, fail, flag = option_part(config["option2"], failed_files, config)
        if flag:
            logging.error("The second option failed, fallback to the pass OCR.")
            done_files.extend([file for file in success if file != ""])
            failed_files = [file['path'] for file in fail]
            success, fail, flag = option_part("pass", failed_files, config)
            done_files.extend([file for file in success if file != ""])
        else:
            done_files.extend(success)
    else:
        done_files.extend(success)
    return done_files



from pdfdeal import Doc2X
from pdfdeal import get_files
import os


def test_single_pdf2file():
    client = Doc2X()
    filepath, _, _ = client.pdf2file(
        pdf_file="tests/pdf/sample.pdf",
        output_path="./Output/test/single/pdf2file",
        output_names=["sample1.docx"],
        output_format="docx",
    )
    if filepath[0] != "":
        assert os.path.exists(filepath[0])
        assert os.path.isfile(filepath[0])
        assert filepath[0].endswith(".docx")
        assert os.path.basename(filepath[0]) == "sample1.docx"


def test_multiple_pdf2file():
    client = Doc2X()
    file_list, rename = get_files("tests/pdf", "pdf", "docx")
    success, failed, flag = client.pdf2file(
        pdf_file=file_list,
        output_path="./Output/test/multiple/pdf2file",
        output_names=rename,
        output_format="docx",
    )
    assert flag
    assert len(success) == len(failed) == 3
    for s in success:
        if s != "":
            assert s.endswith("sample.docx") or s.endswith("sampleB.docx")
    for f in failed:
        if f["path"] != "":
            assert f["path"].endswith("sample_bad.pdf")


# def test_multiple_high_rpm():
#     client = Doc2X()
#     file_list = ["tests/pdf/sample.pdf" for _ in range(20)]
#     success, failed, flag = client.pdf2file(
#         pdf_file=file_list,
#         output_path="./Output/test/high_rpm/pdf2file",
#     )
#     assert len(success) == len(failed) == 20
#     i = 0
#     for s in success:
#         if s != "":
#             assert os.path.exists(s)
#             assert s.endswith(".zip")
#         else:
#             i += 1
#     print(f"===Failed {i} times===")


# def test_translate_pdf():
#     client_personal = Doc2X(apikey=os.getenv("DOC2X_APIKEY_PERSONAL"))
#     file_list = gen_folder_list("tests/pdf", "pdf")
#     success, failed, flag = client_personal.pdf_translate(file_list)
#     assert len(success) == len(failed) == 2
#     if success[0] != "":
#         assert success[0]["texts"]
#         assert success[0]["location"]
#     assert failed[1]["path"].endswith("sample_bad.pdf")



from enum import Enum


class OutputFormat(str, Enum):
    DOCX = "docx"
    TXTS = "texts"
    LATEX = "latex"
    MD = "md"
    MD_DOLLAR = "md_dollar"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )


class RAG_OutputType(str, Enum):
    PDF = "pdf"
    MD = "md"
    TEXTS = "texts"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )


class Support_File_Type(str, Enum):
    PDF = "pdf"
    IMG = "img"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )


class Translate_Model(str, Enum):
    DEEPSEEK = "deepseek"
    GLM = "glm4"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )


class Translate_Language(str, Enum):
    EN = "en"
    ZH = "zh"
    JA = "ja"
    FR = "fr"
    RU = "ru"
    PT = "pt"
    ES = "es"
    DE = "de"
    KO = "ko"
    AR = "ar"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, must be one of {', '.join([m.value for m in cls])}"
        )



import os
import curses

from typing import Tuple
from ..FileTools.ocr import BUILD_IN_OCR
from ..FileTools.tool import BUILD_IN_TOOL
from .i18n import LANGUAGES, WORDS, WORDS_LOCAL
from .connect import BUILD_IN_CONNECT, load_build_in_connect


def init_config(language: str = None) -> dict:
    """This function is used to initialize the configuration of the program."""
    if language is None:
        language = curses_select(LANGUAGES, "Please select the language:")
    words = WORDS[language]

    all_option = [f"OCR: {ocr}" for ocr in BUILD_IN_OCR] + [
        f"Tool: {tool}" for tool in BUILD_IN_TOOL
    ]
    option1 = all_option[curses_select(all_option, words[0])]
    all_option.remove(option1)
    option1 = option1.split(":")[1].strip()
    all_option.append("OCR: pass")
    option2 = all_option[curses_select(all_option, words[1])]
    option2 = option2.split(":")[1].strip()
    Config = {"option1": option1, "option2": option2}

    # If select needs API
    if "doc2x" in option1 or "doc2x" in option2:
        Key, RPM = doc2x_api(words)
        Config["Doc2X_Key"] = Key
        Config["Doc2X_RPM"] = RPM

    return Config


def init_local_config(language: str = None, folder_name=None) -> Tuple[str, dict]:
    """This function is used to initialize the configuration of the local project folder."""
    from .store import get_global_setting

    if language is None:
        language = curses_select(LANGUAGES, "Please select the language:")
    words = WORDS_LOCAL[language]
    base_folder = os.path.expanduser("~/pdfdeal")
    if folder_name is None:
        while True:
            folder_name = input(words[0])
            folder = os.path.join(base_folder, folder_name)
            if os.path.exists(folder):
                print(words[1])
            else:
                break
    try:
        config = get_global_setting()
        if curses_select(["Yes", "No"], words[3]) == 1:
            raise FileNotFoundError
    except FileNotFoundError:
        config = init_config(language)
    print(words[2])
    print(folder)
    os.makedirs(folder)
    connect = curses_select(BUILD_IN_CONNECT, words[4])
    config_init_fnc = load_build_in_connect(BUILD_IN_CONNECT[connect])
    _, config_fnc = config_init_fnc()
    connect_config = config_fnc()
    config.update(connect_config)
    return folder_name, config


def curses_select(selects: list, show: str) -> int:
    stdscr = curses.initscr()
    curses.noecho()
    curses.cbreak()
    stdscr.keypad(True)
    try:
        current_row = 0
        while True:
            stdscr.clear()
            h, w = stdscr.getmaxyx()
            stdscr.addstr(0, 0, show[: w - 1])
            for idx, item in enumerate(selects):
                x = max(0, w // 4 - len(item) // 2)
                y = h // 4 + idx
                if y >= h:
                    break
                item = item[: w - x - 1]
                if idx == current_row:
                    stdscr.attron(curses.A_REVERSE)
                    stdscr.addstr(y, x, item)
                    stdscr.attroff(curses.A_REVERSE)
                else:
                    stdscr.addstr(y, x, item)
            key = stdscr.getch()
            if key == curses.KEY_UP and current_row > 0:
                current_row -= 1
            elif key == curses.KEY_DOWN and current_row < len(selects) - 1:
                current_row += 1
            elif key == curses.KEY_ENTER or key in [10, 13]:
                return current_row
            elif key == curses.KEY_UP and current_row == 0:
                current_row = len(selects) - 1
            elif key == curses.KEY_DOWN and current_row == len(selects) - 1:
                current_row = 0
    finally:
        curses.nocbreak()
        stdscr.keypad(False)
        curses.echo()
        curses.endwin()


def doc2x_api(words: list) -> Tuple[str, int]:
    """This function is used to initialize the API key of doc2x."""
    from ..doc2x import Doc2X

    Key = input(words[2])
    try:
        Doc2X(Key)
    except Exception as e:
        raise Exception(f"{words[3]}:\n {e}")
    RPM = input(words[4])
    assert RPM.isdigit() or RPM == "A" or RPM == "a", "The input is invalid."
    if RPM == "A" or RPM == "a":
        if Key.startswith("sk-"):
            RPM = 10
        else:
            RPM = 1
    return Key, int(RPM)
